import os
import shutil
from pathlib import Path
from typing import Literal, Optional, Set, Tuple, Union

import sqlalchemy as sa
import sqlmodel as sqm
from cloudpathlib import CloudPath
from pydantic import PostgresDsn
from sqlalchemy.future import Engine

from lndb._storage import Storage

from ._exclusion import empty_locker, get_locker
from ._settings_save import save_instance_settings
from ._settings_store import current_instance_settings_file, instance_settings_file

# leave commented out until we understand more how to deal with
# migrations in redun
# https://stackoverflow.com/questions/2614984/sqlite-sqlalchemy-how-to-enforce-foreign-keys
# foreign key constraints for sqlite3
# from sqlite3 import Connection as SQLite3Connection
# from sqlalchemy import event
# from sqlalchemy.engine import Engine
# @event.listens_for(Engine, "connect")
# def _set_sqlite_pragma(dbapi_connection, connection_record):
#     if isinstance(dbapi_connection, SQLite3Connection):
#         cursor = dbapi_connection.cursor()
#         cursor.execute("PRAGMA foreign_keys=ON;")
#         cursor.close()


class InstanceSettings:
    """Instance settings."""

    def __init__(
        self,
        owner: str,  # owner handle
        name: str,  # instance name
        storage_root: Union[str, Path, CloudPath],  # storage location on cloud
        storage_region: Optional[str] = None,
        db: Optional[PostgresDsn] = None,  # DB URI
        schema: Optional[str] = None,  # comma-separated string of schema names
    ):
        self._owner: str = owner
        self._name: str = name
        self._storage: Storage = Storage(storage_root, region=storage_region)
        self._db: Optional[str] = db
        self._schema_str: Optional[str] = schema
        self._engine: Engine = sqm.create_engine(self.db)

    def __repr__(self):
        """Rich string representation."""
        representation = f"Current instance: {self.identifier}"
        attrs = ["owner", "name", "storage", "db", "schema"]
        for attr in attrs:
            value = getattr(self, attr)
            if attr == "storage":
                representation += f"\n- storage root: {value.root}"
                representation += f"\n- storage region: {value.region}"
            else:
                representation += f"\n- {attr}: {value}"
        return representation

    @property
    def owner(self) -> str:
        """Instance owner. A user or organization account handle."""
        return self._owner

    @property
    def name(self) -> str:
        """Instance name."""
        return self._name

    @property
    def identifier(self) -> str:
        """Unique identifier.

        See remote instances at https://lamin.ai/owner/name.
        """
        return f"{self.owner}/{self.name}"

    @property
    def schema(self) -> Set[str]:
        """Schema modules in addition to core schema."""
        if self._schema_str is None:
            return {}  # type: ignore
        else:
            return {schema for schema in self._schema_str.split(",") if schema != ""}

    @property
    def _sqlite_file(self) -> Union[Path, CloudPath]:
        """SQLite file.

        Is a CloudPath if on S3 or GS, otherwise a Path.
        """
        return self.storage.key_to_filepath(f"{self.name}.lndb")

    @property
    def _sqlite_file_local(self) -> Path:
        """Local SQLite file."""
        return self.storage.cloud_to_local_no_update(self._sqlite_file)

    def _update_cloud_sqlite_file(self) -> None:
        """Upload the local sqlite file to the cloud file."""
        if self.is_cloud_sqlite:
            sqlite_file = self._sqlite_file
            cache_file = self.storage.cloud_to_local_no_update(sqlite_file)
            sqlite_file.upload_from(cache_file, force_overwrite_to_cloud=True)  # type: ignore  # noqa
            # doing semi-manually to replace cloudpahlib easily in the future
            cloud_mtime = sqlite_file.stat().st_mtime  # type: ignore
            # this seems to work even if there is an open connection
            # to the cache file
            os.utime(cache_file, times=(cloud_mtime, cloud_mtime))

    def _update_local_sqlite_file(self) -> None:
        """Download the cloud sqlite file if it is newer than local."""
        if self.is_cloud_sqlite:
            sqlite_file = self._sqlite_file
            cache_file = self.storage.cloud_to_local_no_update(sqlite_file)
            # checking cloud mtime several times here because of potential changes
            # during the synchronizization process. Maybe better
            # to make these checks dependent on lock,
            # i.e. if locked check cloud mtime only once.
            if not cache_file.exists():
                cache_file.parent.mkdir(parents=True, exist_ok=True)
                cloud_mtime = sqlite_file.stat().st_mtime  # type: ignore
                sqlite_file.download_to(cache_file)  # type: ignore
                os.utime(cache_file, times=(cloud_mtime, cloud_mtime))
            elif sqlite_file.stat().st_mtime > cache_file.stat().st_mtime:  # type: ignore  # noqa
                # checking the time again because
                # it could be changed in the meantime
                # maybe remove this checks when locked
                cloud_mtime = sqlite_file.stat().st_mtime  # type: ignore
                sqlite_file.download_to(cache_file)  # type: ignore
                os.utime(cache_file, times=(cloud_mtime, cloud_mtime))

    @property
    def db(self) -> str:
        """Database URL."""
        if self._db is None:
            # here, we want the updated sqlite file
            # hence, we don't use self._sqlite_file_local()
            return f"sqlite:///{self.storage.cloud_to_local(self._sqlite_file)}"
        else:
            return self._db

    @property
    def dialect(self) -> Literal["sqlite", "postgresql"]:
        """SQL dialect."""
        if self._db is None or self._db.startswith("sqlite://"):
            return "sqlite"
        else:
            assert self._db.startswith("postgresql")
            return "postgresql"

    @property
    def engine(self) -> Engine:
        """Database engine."""
        return self._engine

    @property
    def is_cloud_sqlite(self) -> bool:
        """Is this a cloud instance with sqlite db."""
        return self.dialect == "sqlite" and self.storage.is_cloud

    @property
    def _cloud_sqlite_locker(self):
        if self.is_cloud_sqlite:
            return get_locker()
        else:
            return empty_locker

    def session(self) -> sqm.Session:
        """Database session."""
        self._update_local_sqlite_file()

        return sqm.Session(self.engine, expire_on_commit=False)

    @property
    def storage(self) -> Storage:
        """Low-level access to storage location."""
        return self._storage

    @property
    def is_remote(self) -> bool:
        """Boolean indicating if an instance has no local component."""
        if not self.storage.is_cloud:
            return False

        def is_local_uri(uri: str):
            if "@localhost:" in uri:
                return True
            if "@0.0.0.0:" in uri:
                return True
            if "@127.0.0.1" in uri:
                return True
            return False

        if self.dialect == "postgresql":
            if is_local_uri(self.db):
                return False
        # returns True for remote SQLite
        # and remote postgres
        return True

    def _persist(self) -> None:
        assert self.name is not None
        if self.storage.type == "local":
            self.storage.root.mkdir(parents=True, exist_ok=True)

        filepath = instance_settings_file(self.name, self.owner)
        # persist under filepath for later reference
        save_instance_settings(self, filepath)
        # persist under current file for auto load
        shutil.copy2(filepath, current_instance_settings_file())
        # persist under settings class for same session reference
        # need to import here to avoid circular import
        from ._settings import settings

        settings._instance_settings = self

    def _is_db_setup(self) -> Tuple[bool, str]:
        """Is the database available and initialized as LaminDB?"""
        if self.dialect == "sqlite":
            if not self._sqlite_file.exists():
                return False, "SQLite file does not exist"
            else:
                return True, ""
        else:  # postgres
            assert self.dialect == "postgresql"
            with self.engine.connect() as conn:
                results = conn.execute(
                    sa.text(
                        """
                    SELECT EXISTS (
                        SELECT FROM
                            information_schema.tables
                        WHERE
                            table_schema LIKE 'public' AND
                            table_name = 'version_yvzi'
                    );
                """
                    )
                ).first()  # returns tuple of boolean
                check = results[0]
                if not check:
                    return False, "Postgres does not seem initialized."
                else:
                    return True, ""



from typing import Union

from lamin_logger import logger
from lnhub_rest._signup_signin import sign_in_hub, sign_up_hub
from sqlalchemy import create_engine

from ._db import upsert
from ._schema import schema
from ._settings import settings
from ._settings_load import load_or_create_user_settings, load_user_settings
from ._settings_save import save_user_settings
from ._settings_store import user_settings_file_email, user_settings_file_handle


def signup(email: str) -> Union[str, None]:
    """Sign up user."""
    response = sign_up_hub(email)
    if response == "handle-exists":  # handle already exists
        logger.error("The handle already exists. Please choose a different one.")
        return "handle-exists"
    if response == "user-exists":  # user already exists
        logger.error("User already exists! Please login instead: `lndb login`.")
        return "user-exists"
    user_settings = load_or_create_user_settings()
    user_settings.email = email
    save_user_settings(user_settings)
    user_settings.password = response
    save_user_settings(user_settings)
    return None  # user needs to confirm email now


def load_user(email: str = None, handle: str = None) -> Union[str, None]:
    if email is not None:
        settings_file = user_settings_file_email(email)
    if handle is not None:
        settings_file = user_settings_file_handle(handle)
    if settings_file.exists():
        user_settings = load_user_settings(settings_file)
        save_user_settings(user_settings)  # needed to save to current_user.env
        assert user_settings.email is not None
    else:
        user_settings = load_or_create_user_settings()
        if email is None:
            raise RuntimeError(
                "Use your email for your first login in a compute environment. "
                "After that, you can use your handle."
            )
        user_settings.email = email
        user_settings.handle = handle
        save_user_settings(user_settings)

    from ._settings import settings

    settings._user_settings = None  # this is to refresh a settings instance

    return None


def login(
    user: str,
    *,
    password: Union[str, None] = None,
) -> Union[str, None]:
    """Log in user."""
    if "@" in user:
        email, handle = user, None
    else:
        email, handle = None, user
    load_user(email, handle)

    user_settings = load_or_create_user_settings()

    if password:
        user_settings.password = password

    if user_settings.email is None:
        raise RuntimeError("No stored user email, please call: lndb login {user}")

    if user_settings.password is None:
        raise RuntimeError(
            "No stored user password, please call: lndb login <your-email>"
            " --password <your-password>"
        )

    response = sign_in_hub(
        user_settings.email, user_settings.password, user_settings.handle
    )
    if response == "could-not-login":
        return response
    elif response == "complete-signup":
        return response
    else:
        user_id, user_handle, user_name, access_token = response
    if handle is None:
        logger.info(f"Your handle is {user_handle} and your id is {user_id}.")
    user_settings.id = user_id
    user_settings.handle = user_handle
    user_settings.name = user_name
    user_settings.access_token = access_token
    save_user_settings(user_settings)

    settings._user_settings = None

    # register login of user in instance db
    # (upsert local user record with cloud data)
    if settings._instance_exists:
        # the above if condition is not safe enough
        # users might delete a database but still keeping the
        # current_instance.env settings file around
        # hence, the if condition will pass despite the database
        # having actually been deleted
        # so, let's do another check
        if settings.instance.dialect == "sqlite":
            # let's check whether the sqlite file is actually available
            if not settings.instance._sqlite_file.exists():
                # if the file doesn't exist, there is no need to
                # log in the user
                # hence, simply end log in here
                return None
            # if the file exists but does not have a user table, raise a warning
            if "core.user" not in schema.list_entities():
                logger.warning(
                    f"An SQLite file {settings.instance._sqlite_file} exists but does not have a user table. "  # noqa
                )
                return None
        else:  # let's check whether we can connect to the instance DB
            db = settings.instance.db
            engine = create_engine(db)
            try:
                engine.connect()
            except Exception:
                logger.warning(
                    f"Connection {db} of current instance not reachable. "
                    "Consider closing it: lndb close."
                )
                return None

        upsert.user(
            settings.user.email,
            settings.user.id,
            settings.user.handle,
            settings.user.name,
        )

    return None



import importlib
import io
from pathlib import Path
from subprocess import run

import sqlalchemy as sa
from alembic import command
from alembic.autogenerate.api import AutogenContext
from alembic.autogenerate.render import _render_cmd_body
from pytest_alembic.config import Config
from pytest_alembic.executor import CommandExecutor, ConnectionExecutor
from pytest_alembic.plugin.error import AlembicTestFailure
from pytest_alembic.runner import MigrationContext
from sqlmodel import SQLModel

from lndb._delete import delete
from lndb._init_instance import init


def get_migration_config(schema_package_loc, *, target_metadata=None, **kwargs):
    if target_metadata is None:
        target_metadata = SQLModel.metadata
    target_metadata.naming_convention = {
        "ix": "ix_%(column_0_label)s",
        "uq": "uq_%(table_name)s_%(column_0_name)s",
        "ck": "ck_%(table_name)s_`%(constraint_name)s`",
        "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
        "pk": "pk_%(table_name)s",
    }
    raw_config = dict(
        config_file_name=f"{schema_package_loc}/alembic.ini",
        script_location=f"{schema_package_loc}/migrations",
        target_metadata=target_metadata,
        **kwargs,
    )
    config = Config.from_raw_config(raw_config)
    return config


def get_migration_context(schema_package, db, include_schemas=None):
    engine = sa.create_engine(db)
    config = get_migration_config(schema_package, include_schemas=include_schemas)
    command_executor = CommandExecutor.from_config(config)
    command_executor.configure(connection=engine)
    migration_context = MigrationContext.from_config(
        config, command_executor, ConnectionExecutor(), engine
    )
    return migration_context


def get_migration_id_from_scripts(schema_package):
    config = get_migration_config(schema_package)
    output_buffer = io.StringIO()
    # get the id of the latest migration script
    if Path(f"./{schema_package}/migrations/versions").exists():
        command.heads(config.make_alembic_config(stdout=output_buffer))
        output = output_buffer.getvalue()
        migration_id = output.split(" ")[0]
    else:  # there is no scripts directory
        migration_id = ""
    return migration_id


def migration_id_is_consistent(schema_package):
    migration_id = get_migration_id_from_scripts(schema_package)
    package = importlib.import_module(schema_package)
    if package._migration is None:
        manual_migration_id = ""
    else:
        manual_migration_id = package._migration
    return manual_migration_id == migration_id


def model_definitions_match_ddl(schema_package, db=None, dialect_name="sqlite"):
    if db is None and dialect_name == "sqlite":
        db = "sqlite:///testdb/testdb.lndb"
        # need to call init to reload schema
        init(storage="testdb", _migrate=False)
    elif db is None and dialect_name == "postgresql":
        # requires postgres has been set up through _nox_tools
        db = "postgresql://postgres:pwd@0.0.0.0:5432/pgtest"
        # need to call init to reload schema
        init(db=db, storage="pgtest", _migrate=False)
    elif db is None:
        raise NotImplementedError(
            "Only sqlite and postgres test databases are implemented."
        )
    # the below is for debugging purposes, something with indexes doesn't work 100%
    # e = sa.create_engine(db)
    # print(sa.inspect(e).get_indexes("core.dobject"))
    include_schemas = True if dialect_name == "postgresql" else False
    migration_context = get_migration_context(
        schema_package, db, include_schemas=include_schemas
    )
    execute_model_definitions_match_ddl(migration_context)
    if dialect_name == "postgresql":
        run("docker stop pgtest && docker rm pgtest", shell=True)
        delete("pgtest")
    else:
        delete("testdb")


def execute_model_definitions_match_ddl(alembic_runner):
    # this function is largely copied from the
    # MIT licensed https://github.com/schireson/pytest-alembic
    """Assert that the state of the migrations matches the state of the models.

    In general, the set of migrations in the history should coalesce into DDL
    which is described by the current set of models. Therefore, a call to
    `revision --autogenerate` should always generate an empty migration (e.g.
    find no difference between your database (i.e. migrations history) and your
    models).
    """

    def verify_is_empty_revision(migration_context, __, directives):
        script = directives[0]

        migration_is_empty = script.upgrade_ops.is_empty()
        if not migration_is_empty:
            autogen_context = AutogenContext(migration_context)
            rendered_upgrade = _render_cmd_body(script.upgrade_ops, autogen_context)

            if not migration_is_empty:
                raise AlembicTestFailure(
                    "The models describing the DDL of your database are out of sync"
                    " with the set of steps described in the revision history. This"
                    " usually means that someone has made manual changes to the"
                    " database's DDL, or some model has been changed without also"
                    " generating a migration to describe that"
                    f" change.\n{rendered_upgrade}"
                )

    try:
        alembic_runner.migrate_up_to("heads")
    except RuntimeError as e:
        raise AlembicTestFailure(
            "Failed to upgrade to the head revision. This means the historical chain"
            f" from an empty database, to the current revision is not possible.\n{e}"
        )

    alembic_runner.generate_revision(
        message="test revision",
        autogenerate=True,
        prevent_file_generation=True,
        process_revision_directives=verify_is_empty_revision,
    )



from pathlib import Path
from typing import Dict

import pandas as pd
import sqlmodel as sqm
from IPython.display import display as ipython_display

from ._db import insert
from ._settings_instance import InstanceSettings


def write_bionty_versions(isettings: InstanceSettings):
    """Write bionty _current.yaml to the CurrentBiontyVersions table."""
    if "bionty" in isettings.schema:
        import bionty as bt
        from bionty.dev._io import load_yaml
        from lnschema_bionty import dev

        basedir = Path(bt.__file__).parent / "versions"
        _current = load_yaml(basedir / "_current.yaml")
        _local = load_yaml(basedir / "_local.yaml")

        # here we set integer ids from 0
        records = []
        for i, (entity, db) in enumerate(_current.items()):
            db_name = next(iter(db))
            db_version = db[db_name]
            record = dev.BiontyVersions(
                id=i,
                entity=entity,
                database=db_name,
                database_v=str(db_version),
                database_url=_local.get(entity)
                .get(db_name)
                .get("versions")
                .get(db_version),
            )
            records.append(record)

        insert.bionty_versions(records)


def load_bionty_versions(isettings: InstanceSettings, display: bool = False):
    """Write CurrentBiontyVersions to _lndb.yaml in bionty."""
    if "bionty" in isettings.schema:
        import bionty as bt
        from bionty.dev._io import write_yaml
        from lnschema_bionty import dev

        basedir = Path(bt.__file__).parent / "versions"

        stmt = sqm.select(dev.BiontyVersions).join(dev.CurrentBiontyVersions)
        with isettings.session() as ss:
            results = ss.exec(stmt).all()
        # avoid breaking change
        # if no versions were written in the db, write versions from bionty
        if len(results) == 0:
            write_bionty_versions(isettings)
        records = [row.dict() for row in results]
        df = pd.DataFrame.from_records(records)
        df_lndb = df.set_index("entity")[["database", "database_v"]]
        if display:
            ipython_display(df_lndb)
        lndb_dict: Dict = {}
        for entity, db in df_lndb.iterrows():
            lndb_dict[entity] = {}
            lndb_dict[entity][db["database"]] = db["database_v"]
        write_yaml(lndb_dict, basedir / "_lndb.yaml")



from lnhub_rest import __version__ as lnhub_rest_v
from packaging import version

if version.parse(lnhub_rest_v) != version.parse("0.3.0"):
    raise RuntimeError("Upgrade lnhub_rest: pip install lnhub_rest==0.3.0")



from ._settings_store import current_instance_settings_file


def close() -> None:
    """Close existing instance.

    Returns `None` if succeeds, otherwise a string error code.
    """
    current_instance_settings_file().unlink()



from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config, pool
from sqlmodel import SQLModel

config = context.config

if config.config_file_name is not None:
    fileConfig(config.config_file_name, disable_existing_loggers=False)

target_metadata = SQLModel.metadata
target_metadata.naming_convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_`%(constraint_name)s`",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}

_schema_id = None
# from {package_name} import *  # noqa
# from {package_name} import _schema_id  # noqa


def _upg_dwg_token(db: str):
    if db != "sqlite:///testdb/testdb.lndb":
        return {"upgrade_token": "", "downgrade_token": ""}
    else:
        return {}


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    add_kwargs = {}
    if url.startswith("sqlite:///"):
        add_kwargs = _upg_dwg_token(url)
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        **add_kwargs,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    config_section = config.get_section(config.config_ini_section)
    # see https://pytest-alembic.readthedocs.io/en/latest/setup.html#env-py
    connectable = context.config.attributes.get("connection", None)
    # below follows the standard case
    if connectable is None:
        connectable = engine_from_config(
            config_section,
            prefix="sqlalchemy.",
            poolclass=pool.NullPool,
        )

    add_kwargs = {}
    if "sqlalchemy.url" in config_section:
        if config_section["sqlalchemy.url"].startswith("sqlite:///"):
            render_as_batch = True
            # for users that rely on CI to generate migration script
            # we generate empty script here
            add_kwargs = _upg_dwg_token(config_section["sqlalchemy.url"])
    else:
        render_as_batch = False

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            version_table=f"migration_{_schema_id}",
            render_as_batch=render_as_batch,
            compare_type=True,
            include_schemas=True,
            **add_kwargs,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



from pathlib import Path

import nox

nox.options.reuse_existing_virtualenvs = True


@nox.session
def lint(session: nox.Session) -> None:
    session.install("pre-commit")
    session.run("pre-commit", "install")
    session.run("pre-commit", "run", "--all-files")


@nox.session(python=["3.7", "3.8", "3.9", "3.10", "3.11"])
def build(session):
    session.install(".[dev,test]")
    login_user_1 = "lndb login testuser1@lamin.ai --password cEvcwMJFX4OwbsYVaMt2Os6GxxGgDUlBGILs2RyS"  # noqa
    session.run(*(login_user_1.split(" ")))
    login_user_2 = "lndb login testuser2@lamin.ai --password goeoNJKE61ygbz1vhaCVynGERaRrlviPBVQsjkhz"  # noqa
    session.run(*(login_user_2.split(" ")), external=True)
    session.run(
        "pytest",
        "-s",
        "--cov=lndb_setup",
        "--cov-append",
        "--cov-report=term-missing",
    )
    session.run("coverage", "xml")
    prefix = "." if Path("./lndocs").exists() else ".."
    session.install(f"{prefix}/lndocs")
    session.run("lndocs")



"""Clone a database to a test database, with built in fetch-depth."""
# This starts out with https://stackoverflow.com/questions/70392123
import time
from pathlib import Path
from subprocess import run
from typing import Optional

import sqlalchemy as sa
from lamin_logger import logger
from sqlalchemy import MetaData, create_engine, func, select

from ._settings import settings
from ._settings_instance import InstanceSettings


def setup_local_test_sqlite_file(
    src_settings: InstanceSettings, return_dir: bool = False
):
    path = src_settings._sqlite_file_local
    new_stem = path.stem + "_test"
    tgt_sqlite_dir = Path.cwd() / new_stem
    if return_dir:
        return tgt_sqlite_dir
    tgt_sqlite_file = tgt_sqlite_dir / f"{new_stem}{path.suffix}"
    tgt_sqlite_file.parent.mkdir(exist_ok=True)
    if tgt_sqlite_file.exists():
        tgt_sqlite_file.unlink()
    tgt_db = f"sqlite:///{tgt_sqlite_file}"
    return tgt_db


def setup_local_test_postgres(name: str = "pgtest"):
    process = run(
        f"docker run --name {name} -e POSTGRES_PASSWORD=pwd"
        f" -e POSTGRES_DB={name} -d -p 5432:5432 postgres",  # noqa
        shell=True,
    )
    if process.returncode == 0:
        logger.info(
            f"Created Postgres test instance. It runs in docker container '{name}'."
        )
    else:
        raise RuntimeError("Failed to set up postgres test instance.")
    time.sleep(2)
    return f"postgresql://postgres:pwd@0.0.0.0:5432/{name}"


def clone_schema(
    schema, src_conn, src_metadata, tgt_conn, tgt_metadata, tgt_engine, depth: int
):
    # switch off foreign key integrity
    if src_conn.dialect.name == "postgresql":
        tgt_conn.execute(sa.sql.text("SET session_replication_role = replica;"))

    # create all tables in target database
    for table in src_metadata.sorted_tables:
        if not sa.inspect(tgt_engine).has_table(table.name, table.schema):
            table.create(bind=tgt_engine)
    # refresh metadata before copying data
    tgt_metadata.clear()
    tgt_metadata.reflect(bind=tgt_engine, schema=schema)

    # copy data
    print("Cloning: ", end="")
    for table in tgt_metadata.sorted_tables:
        if table.schema != schema:
            continue
        src_table = src_metadata.tables[f"{schema}.{table.name}"]
        n_rows = -1  # indicates no rows
        if len(list(src_table.primary_key)) > 0:
            pk_col = getattr(src_table.c, list(src_table.primary_key)[0].name)
            n_rows = int(
                src_conn.execute(
                    select([func.count(pk_col)]).select_from(src_table)
                ).scalar()
            )
        print(f"{table.name} ({n_rows})", end=", ")
        offset = max(n_rows - depth, 0)
        rows = src_conn.execute(src_table.select().offset(offset))
        values = [row._asdict() for index, row in enumerate(rows)]
        if len(values) > 0:
            tgt_conn.execute(table.insert(), values)
            tgt_conn.commit()


def clone_test(src_settings: Optional[InstanceSettings] = None, depth: int = 10):
    """Clone from current instance to a test instance."""
    if src_settings is None:
        src_settings = settings.instance
    src_engine = create_engine(src_settings.db)
    src_metadata = MetaData()

    if src_settings.dialect == "sqlite":
        tgt_db = setup_local_test_sqlite_file(src_settings)
    else:
        tgt_db = setup_local_test_postgres()

    assert tgt_db != src_settings.db

    tgt_engine = create_engine(tgt_db, future=True)
    tgt_metadata = MetaData()
    src_conn = src_engine.connect()
    src_metadata.reflect(bind=src_engine)
    tgt_conn = tgt_engine.connect()

    # create all schemas in target database
    src_schemas = src_conn.dialect.get_schema_names(src_conn)
    for schemaname in src_schemas:
        if schemaname not in tgt_conn.dialect.get_schema_names(tgt_conn):
            tgt_conn.execute(sa.schema.CreateSchema(schemaname))
        tgt_conn.commit()

    # only relevant for postgres, clean out some default schemas
    for schema in [
        "information_schema",
        "auth",
        "graphql",
        "graphql_public",
        "realtime",
        "extensions",
        "storage",
    ]:
        if schema in src_schemas:
            src_schemas.remove(schema)

    for schema in src_schemas:
        src_metadata.reflect(bind=src_engine, schema=schema)
        tgt_metadata.reflect(bind=tgt_engine, schema=schema)
        if src_engine.dialect.name != "sqlite":
            print(f"\nSchema: {schema}")
        clone_schema(
            schema, src_conn, src_metadata, tgt_conn, tgt_metadata, tgt_engine, depth
        )

    return tgt_db
